﻿// Agent.cs
// This file containts the Agent class, which represents a man/woman


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace UnbalancedMatchSim
{
    /// <summary>
    /// A class representing a man/woman
    /// </summary>
    public class Agent
    {
        public const int UNMATCHED_INDICATOR = -1;

        public int ID; // this ID will be the agent's identifier

        private int K; // number of potential spouse == MAX_RANK + 1
        private int[] prefList; // preferences over the other side, giving indexes
        private int assignmentRank; // rank of current assigned partner, lower=better

        private int maxAcceptableRank; // maximal acceptable rank

        /// <summary>
        /// This function creates an agent with blank (illegal) preferences.
        /// </summary>
        /// <param name="K">Number of colleges</param>
        /// <param name="id">ID of student</param>
        public Agent(int K, int id)
        {
            ID = id;
            this.K = K;
            prefList = new int[K];

            // init by making the agent unmatched
            assignmentRank = K + 1;
            maxAcceptableRank = K;

        }


        /// <summary>
        /// Draws preferences for the student and scores.
        /// Pref are drawn uniformly at random. 
        /// </summary>
        public void DrawPref()
        {
            // draw the student's preferences over colleges
            prefList = PrefGenerator.getPermutation(K);
        }

        /// <summary>
        /// Draws preferences using the given utility function and characteristics.
        /// </summary>
        /// <param name="utilFunction"></param>
        /// <param name="characteristicsSpouses"></param>
        /// <param name="mySideChar"></param>
        /// <param name="isMen"></param>
        public void DrawPref(IUtility utilFunction, ICharacteristics characteristicsSpouses, ICharacteristics mySideChar, bool isMen)
        {
            double[] u = new double[K];
            int[] ind = new int[K];

            int acceptableCount = 0; 

            double unmatchedUtil = utilFunction.getUnmatchedUtil(this.ID, characteristicsSpouses, mySideChar, isMen);

            for (int i = 0; i < K; i++)
            {
                double uDraw = utilFunction.drawUtil(this.ID, i, characteristicsSpouses, mySideChar, isMen);

                if (uDraw > unmatchedUtil){
                    ind[acceptableCount] = i;
                    // taking negative of the utility, 
                    // so the sorting below will give the highest utility first
                    u[i] = -1 * uDraw;
                    acceptableCount++;
                }
            }

            Array.Sort(u, ind,0,acceptableCount);
            
            // make a new array with only the acceptable spouses, followed by the marker for unmatched.
            prefList = new int[acceptableCount + 1];

            Array.Copy(ind, prefList, acceptableCount);
            prefList[acceptableCount] = UNMATCHED_INDICATOR;

            maxAcceptableRank = acceptableCount;
            // prefList = ind;


        }
        

        /// <summary>
        ///  returns the current assigned partner
        /// </summary>
        /// <returns></returns>
        public int getAssignedSpouse()
        {
            if (!isUnmatched())
                return prefList[assignmentRank];
            else
                return Agent.UNMATCHED_INDICATOR;
        }

        public Boolean isUnmatched()
        {
            return assignmentRank < 0 || assignmentRank >= maxAcceptableRank;
        }

        /// <summary>
        /// Considers a proposal from an agent and returns the ID of the rejected agent
        /// (which may be the proposer), or -1 if the agent was previously unmatched.
        /// </summary>
        /// <param name="proposingAgentID"></param>
        public int makeProposal(int proposingAgentID)
        {
            int rank = Array.IndexOf<int>(prefList, proposingAgentID);
            // If the proposing agent is not acceptable, reject him
            if (rank >= maxAcceptableRank || rank == -1 )
            {
                return proposingAgentID;
            }
            // else, check if the woman is currently matched to a more preferred man
            if (isUnmatched())
            {
                assignmentRank = rank;
                return Agent.UNMATCHED_INDICATOR;
            }
            else
            {
                if (assignmentRank < rank)
                    return proposingAgentID;
                else
                {
                    int rejectedAgentID = getAssignedSpouse();
                    assignmentRank = rank;
                    return rejectedAgentID;
                }
            }
        }

        public int getRank(int spouseID)
        {
            return Array.IndexOf(prefList, spouseID);
        }


        public void setUnassigned()
        {
            assignmentRank = K + 1;
        }

        public void setTopSpouse()
        {
            assignmentRank = 0;
        }

        public void reject()
        {
            assignmentRank++;
        }

        public override string ToString()
        {
            return ID.ToString();
        }

        public string ToStringWithPref()
        {
            string s = ToString() + "\t:";
            for (int i = 0; i < maxAcceptableRank; i++)
            {
                s = s + " " + prefList[i];
            }
            return s;
        }

    }
}
