﻿//Economy.cs
// Includes the Economy class, which represents a collection of men and women.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace UnbalancedMatchSim
{

    /// <summary>
    /// A class that represents an economy, being a collection of men and women.
    /// Includes functionality for drawing preferences and calculating stats. 
    /// </summary>
    class Economy
    {

        Agent[] men;
        Agent[] women;

        int numMen;
        int numWomen;

        // Drawn Characteristics

        ICharacteristics menChar = null;
        ICharacteristics womenChar = null;

        //variable to store the stats of the economy
        // (after MOSM and WOSM are calculated) 

        int[] muMOSM;
        int[] muWOSM;

        public int totalRankMOSM = 0;
        public int totalRankWOSM = 0;
        public int numMatched = 0;
        public int numMult = 0;

       
        /// <summary>
        /// Initiates a new economy with the set number of men and women. Preferences are not initialized. 
        /// </summary>
        /// <param name="numMen"></param>
        /// <param name="numWomen"></param>
        public Economy(int numMen, int numWomen)
        {

            this.numMen = numMen;
            this.numWomen = numWomen;

            men = new Agent[numMen];
            women = new Agent[numWomen];

            for (int j = 0; j < numWomen; j++)
                women[j] = new Agent(numMen, j);
            for (int i = 0; i < numMen; i++)
                men[i] = new Agent(numWomen, i);

        }

        /// <summary>
        /// Draws Uniform IID preferences for all agents in the economy
        /// </summary>
        public void drawUniformIIDPref()
        {
            foreach (Agent w in women)
                w.DrawPref();
            foreach (Agent m in men)
                m.DrawPref();
        }

        

        /// <summary>
        /// Draw preferences for the entire economy from the given utility function
        /// (characteristics are also drawn in the process) 
        /// </summary>
        /// <param name="utility"></param>
        public void drawPrefFromUtility(IUtility utility)
        {
            menChar = utility.drawChar(numMen, true);
            womenChar = utility.drawChar(numWomen, false);

            foreach (Agent m in men)
            {
                m.DrawPref(utility, womenChar, menChar, true);
            }
            foreach (Agent w in women)
            {
                w.DrawPref(utility, menChar, womenChar, false);
            }
        }


        /// <summary>
        /// Runs MPDA for the economy and calculates totalRankMOSM and NumMatched
        /// </summary>
        public void calcMOSM()
        {
            muMOSM = UnequalMatching.MPDA(men, women);

            numMatched = 0;
            totalRankMOSM = 0;

            foreach (Agent m in men)
            {
                if (!m.isUnmatched())
                {
                    numMatched++;
                    totalRankMOSM += m.getRank(m.getAssignedSpouse());
                }
            }
        }


        /// <summary>
        /// Runs WPDA for the economy and calculates totalRankWOSM 
        /// </summary>
        public void calcWOSM()
        {
            UnequalMatching.MPDA(women, men);
            totalRankWOSM = 0;
            muWOSM = new int[numMen];

            foreach (Agent m in men)
            {
                if (!m.isUnmatched())
                {
                    totalRankWOSM += m.getRank(m.getAssignedSpouse());
                }
                muWOSM[m.ID] = m.getAssignedSpouse();
            }
        }

    
        /// <summary>
        /// Calculates the number of men who have different wifes under the WOSM and MOSM. 
        /// Assumes that calcWOSM() and calcMOSM() were already run. 
        /// </summary>
        public void calcMult()
        {
            numMult = 0;

            for (int i = 0; i < numMen; i++)
            {
                if (muMOSM[i] != muWOSM[i])
                    numMult++;
            }
        }

        /// <summary>
        /// Caculates all the stats for the economy. 
        /// </summary>
        public void calcMatchingStats()
        {
            calcMOSM();
            calcWOSM();
            calcMult();
        }


        /// <summary>
        /// Prints the economy for debug purpuses.
        /// </summary>
        /// <param name="sw"></param>
        public void printEconomy(StreamWriter sw)
        {
            sw.WriteLine();
            sw.WriteLine(" Women: ");
            for (int j = 0; j < numWomen; j++)
                sw.Write("w" + women[j].ToStringWithPref() + "\n");
            if (null != womenChar)           
                womenChar.printCharac(sw);

            sw.WriteLine(" Men: ");
            for (int i = 0; i < numMen; i++)
                sw.Write("m" + men[i].ToStringWithPref() + "\n");
            if (null != menChar)
                menChar.printCharac(sw);

            sw.WriteLine();

        }


    }
}
