﻿// Program.cs
// Contains the Main(string[] args) and functions for setting parameters and outputing simulation results. 



using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UnbalancedMatchSim.Utilities;





namespace UnbalancedMatchSim
{
    class Program
    {
        
        /// Configuration Parameters 
        ///////////////////////////

        // Output folder
        // -------------
        // The program will crach if this folder does not exist.
        public const string OUTPUT_FOLDER = "../../../Output/";

        // Iteration Num
        // -------------
        // Number of iterations for each market configuration
        public const int ITERATION_NUM = 10;
        
        // Full Stats Flag
        // ---------------
        // If this is set to false the program reports only averages.
        // If it is set to true it will report avg, std, 5 percentile, and 95 percentile
        public const bool IS_PRINT_FULL_STATS = false;





        /// Labels for printing
        ///////////////////////
        
        public const string LABELROW_FULLSTATS = "x1	numMen	x2	numWomen	"
            + RandomTraitsLineUtil.UTIL_LABELROW +
            "x4	Rmsom	stdmosm	95msom	5msom	x5	Rwsom	stdwsom	95wsom	5wsom	x6	pctMult	pctMultstd	pctMult95	pctMult5	x7	pctMatched	pctMatchedstd	pctMatched95	pctMatched5";
        public const string LABELROW_SHORT = "x1	numMen	x2	numWomen	"
            + RandomTraitsLineUtil.UTIL_LABELROW +
            "x4	Rmsom	x5	Rwsom	x6	pctMult	x7	pctMatched";



        /// <summary>
        /// The main file calls the functions for the different simulations
        /// </summary>
        /// <param name="args"></param>
        static void Main(string[] args)
        {
            var watch = Stopwatch.StartNew();
            simUIIDmarkets(ITERATION_NUM, IS_PRINT_FULL_STATS);
            CorrelationShrinkCorr(ITERATION_NUM, IS_PRINT_FULL_STATS);
            ShortSideChoosingCorr(ITERATION_NUM);
            ShortSideChoosingShortLists(ITERATION_NUM, IS_PRINT_FULL_STATS);

            watch.Stop();
            Console.WriteLine(watch.ElapsedMilliseconds);

            Console.WriteLine("Hello");
            Console.ReadKey();
        }




        // Simulation of the model
        ////////////////////////////////////
        ////////////////////////////////////

        /// <summary>
        /// This function simulation markets with uniform IID preferences. 
        /// 
        /// The parameters for the range of simulations to run are hard coded inside the function (see below).
        /// </summary>
        /// <param name="iterationNum"></param>
        /// <param name="isFullStats"></param>
        public static void simUIIDmarkets(int iterationNum, bool isFullStats)
        {
            StreamWriter sw = new StreamWriter(OUTPUT_FOLDER + "UIIDmarkets__IterationNum=" + iterationNum + ".txt", false, Encoding.ASCII);
            sw.AutoFlush = true;
           
            if (isFullStats)
                sw.WriteLine(LABELROW_FULLSTATS);
            else
                sw.WriteLine(LABELROW_SHORT);
   

            // Iterate over markets
            ///////////////

            int numWomen = 40;
            for (int numMen = 20; numMen <= 60; numMen++)
                runAndPrintForParams(null, sw, numMen, numWomen, iterationNum, isFullStats);

            sw.Close();
        }



        // Correlation shrinks the core
        ////////////////////////////////////
        ////////////////////////////////////

        /// <summary>
        /// This function creates different markets with various correlation parameters.
        /// It uses the utlity model:
        /// u_i(j) = \beta * x_j^A - \gamma * (x_i^D - x_j^D)^2 + \varepsilon_{ij}
        /// 
         /// 
        /// The parameters for the range of simulations to run are hard coded inside the function (see below).
        /// </summary>
        /// <param name="iterationNum"></param>
        /// <param name="isFullStats"></param>
        public static void CorrelationShrinkCorr(int iterationNum, bool isFullStats)
        {
            StreamWriter sw = new StreamWriter(OUTPUT_FOLDER + "CorrPct_exp=2__IterationNum=" + iterationNum + ".txt", false, Encoding.ASCII);
            sw.AutoFlush = true;

            if (isFullStats)
            {
                sw.WriteLine(LABELROW_FULLSTATS);
            }
            else
            {
                sw.WriteLine(LABELROW_SHORT);
            }
            
            
            IUtility utilFunction = new RandomTraitsLineUtil();


            // utilFunction.writeUtilityModelDescrition(sw);


            // Paramters
            ///////////////


            int numWomen = 40;
            int[] menVals = { 40, 41, 45, 60 };
            utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_EXPONENET, 2);


            for (double absCorrCoe = 0; absCorrCoe <= 20; absCorrCoe += 1)
            {
                for (double mktCorrDiff = -20; mktCorrDiff <= 20; mktCorrDiff += 1)
                {
                    foreach (int numMen in menVals)
                    {
                        utilFunction.setCoefficient(RandomTraitsLineUtil.COE_ABS, absCorrCoe);
                        utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_MIN, mktCorrDiff);
                        utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_PLUS, mktCorrDiff);
                        runAndPrintForParams(utilFunction, sw, numMen, numWomen, iterationNum, isFullStats);
                    }
                    Console.WriteLine("{0:n} {1:n}", absCorrCoe, mktCorrDiff);
                }
            }
            sw.Close();
        }


        // Negative Alignment and the size of the Core
        ////////////////////////////////////
        ////////////////////////////////////
                
        /// <summary>
        /// This function creates different markets with misalignment between men's and women's preferences. 
        /// It uses the utlity model:
        /// u_i(j) = \beta * x_j^A - \gamma * (x_i^D - x_j^D)^2 + \varepsilon_{ij}
        /// 
        /// The value of \gamma_M for men is made to be the negative of \gamma_W used for women, 
        /// implying that women want to be far (close) while men want to be close (far).
        /// 
        /// The parameters for the range of simulations to run are hard coded inside the function (see below).
        /// </summary>
        /// <param name="iterationNum"></param>
        /// <param name="isFullStats"></param>
        public static void NegativeAlignment(int iterationNum)
        {


            StreamWriter sw = new StreamWriter(OUTPUT_FOLDER + "NegAlignmentRank__ItrNum=" + iterationNum + ".txt", false, Encoding.ASCII);
            sw.AutoFlush = true;
            sw.WriteLine(LABELROW_SHORT);

            IUtility utilFunction = new RandomTraitsLineUtil();

            int numWomen = 40;
            IEnumerable<int> numMenRange = Enumerable.Range(20, 41);
            int[] correlationVals = { 0, 1, 2, 5, 10, 100 };

            utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_EXPONENET, 2);
            utilFunction.setCoefficient(RandomTraitsLineUtil.COE_ABS, 0);

            utilFunction.setParam(RandomTraitsLineUtil.PARAM_IS_DIFF_REVERSED_FOR_WOMEN, RandomTraitsLineUtil.PARAM_TRUE);


            foreach (double diffCorrCoe in correlationVals)
            {
                foreach (int numMen in numMenRange)
                {
                    utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_MIN, diffCorrCoe);
                    utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_PLUS, diffCorrCoe);

                    runAndPrintForParams(utilFunction, sw, numMen, numWomen, iterationNum, false);

                }
                Console.WriteLine("DiffNeg " + diffCorrCoe);
            }
            sw.Close();


            // Make a 3D plot of %mult
            ////////////////
            
            iterationNum = 100;
            sw = new StreamWriter(OUTPUT_FOLDER + "NegativeAlignedCoreSize__ItrNum=" + iterationNum + ".txt", false, Encoding.ASCII);
            sw.WriteLine(LABELROW_SHORT);

            numWomen = 40;
            int[] menVals = {41,42,45,50,60};
            utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_EXPONENET, 2);


            for (double absCorrCoe = 0; absCorrCoe <= 20; absCorrCoe += 1)
            {
                for (double mktCorrDiff = -20; mktCorrDiff <= 20; mktCorrDiff += 1)
                {
                    foreach (int numMen in menVals)
                    {
                        utilFunction.setCoefficient(RandomTraitsLineUtil.COE_ABS, absCorrCoe);
                        utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_MIN, mktCorrDiff);
                        utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_PLUS, mktCorrDiff);
                        runAndPrintForParams(utilFunction, sw, numMen, numWomen, iterationNum, false);

                    }
                    Console.WriteLine("{0:n} {1:n}", absCorrCoe, mktCorrDiff);
                }
            }
            sw.Close();



        }


        // Short side gets to choose, even with some correlation
        ////////////////////////////////////
        ////////////////////////////////////

        /// <summary>
        /// This function creates different markets with various correlation parameters.
        /// It uses the utlity model:
        /// u_i(j) = \beta x_j^A - \gamma^+ (| x_i^D - x_j^D|_+ )^2 - \gamma^- (| x_j^D - x_i^D|_+ )^2 + \varepsilon_{ij}
        /// 
        /// It resports the results to several files, each including a different range of market parameters. 
        /// 
        /// The parameters for the range of simulations to run are hard coded inside the function (see below).
        /// </summary>
        /// </summary>
        /// <param name="iterationNum"></param>
        public static void ShortSideChoosingCorr(int iterationNum)
        {
            IUtility utilFunction = new RandomTraitsLineUtil();

            // Paramters 
            ///////////////////
            int numWomen = 40;
            IEnumerable<int> numMenRange = Enumerable.Range(20, 40);
            int[] correlationVals = { 0, 1, 2, 5, 10, 15, 20, 100, 1000 };

            utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_EXPONENET, 2);



            // plot 1: varying abs correlation
            //////////////////////////////////
            StreamWriter sw = new StreamWriter(OUTPUT_FOLDER + "AbsCorrChoose__ItrNum=" + iterationNum + ".txt", false, Encoding.ASCII);
            sw.AutoFlush = true;
            sw.WriteLine(LABELROW_SHORT);

            utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_MIN, 0);
            utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_PLUS, 0);
            foreach (double absCorrCoe in correlationVals)
            {
                foreach (int numMen in numMenRange)
                {
                    utilFunction.setCoefficient(RandomTraitsLineUtil.COE_ABS, absCorrCoe);
                    runAndPrintForParams(utilFunction, sw, numMen, numWomen, iterationNum, false);

                }
                Console.WriteLine("Abs " + absCorrCoe);
            }
            sw.Close();


            // plot 2: varying diff correlation with aligned pref
            ///////////////////////////////////////////////////
            sw = new StreamWriter(OUTPUT_FOLDER + "AlignedCorrChoose__ItrNum=" + iterationNum + ".txt", false, Encoding.ASCII);
            sw.AutoFlush = true;
            sw.WriteLine(LABELROW_SHORT);

            utilFunction.setCoefficient(RandomTraitsLineUtil.COE_ABS, 0);
            foreach (double diffCorrCoe in correlationVals)
            {
                foreach (int numMen in numMenRange)
                {
                    utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_MIN, -diffCorrCoe);
                    utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_PLUS, -diffCorrCoe);
                    runAndPrintForParams(utilFunction, sw, numMen, numWomen, iterationNum, false);

                }
                Console.WriteLine("DiffPos " + diffCorrCoe);
            }
            sw.Close();

            // plot 4: makign all coeficients equal 
            ///////////////////////////////////////
            sw = new StreamWriter(OUTPUT_FOLDER + "AlleqCorrChoose__ItrNum=" + iterationNum + ".txt", false, Encoding.ASCII);
            sw.AutoFlush = true;
            sw.WriteLine(LABELROW_SHORT);

            foreach (double diffCorrCoe in correlationVals)
            {
                foreach (int numMen in numMenRange)
                {
                    utilFunction.setCoefficient(RandomTraitsLineUtil.COE_ABS, diffCorrCoe);
                    utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_MIN, -diffCorrCoe);
                    utilFunction.setCoefficient(RandomTraitsLineUtil.COE_DIFF_PLUS, -diffCorrCoe);
                    runAndPrintForParams(utilFunction, sw, numMen, numWomen, iterationNum, false);

                }
                Console.WriteLine("AllEq " + diffCorrCoe);
            }

            sw.Close();
        }


        /// Short side chooses with short lists
        //////////////////////////////////////////////
        //////////////////////////////////////////////

        /// <summary>
        /// This function simulates markets where men and women have partial lists.
        /// Each market is given by a number of men and women, as well as two reservation utilities 
        /// for remaining unmatched, one for men and one for women. Preferences are drawn uniformly i.i.d.
        /// </summary>
        /// <param name="iterationNum"></param>
        /// <param name="isFullStats"></param>
        public static void ShortSideChoosingShortLists(int iterationNum, bool isFullStats)
        {

            StreamWriter sw = new StreamWriter(OUTPUT_FOLDER + "PartialLists_itNum=" + iterationNum + ".txt", false, Encoding.ASCII);
            sw.AutoFlush = true;
            sw.WriteLine(LABELROW_SHORT);

            // Paramters
            /////////////////////////

            int numWomen = 40;

            IUtility utilFunction = new RandomTraitsLineUtil();

            // We use the EV distribution for drawing utilities, and therefore the reservation utilities 
            // need to correspond to percentiles of the EV distribution. 
            // 
            // 10% intervals of the EV distribution:
            // .1       .2      .3      .4      .5      .6      .7      .8      .9
            // -0.834	-0.476	-0.186	0.087	0.367	0.672	1.031	1.500	2.250
            
            // utilities are drawn using EV distribution, these correspond to  ~0%,33%, 66% of the disribution
            // (i.e., probability of a spouse being unacceptable = ~0, .33, .66 ) 
            double[] unmatchedUtilRange = { -1000, -0.094, 0.903};
            

            foreach (double unmatchedUmen in unmatchedUtilRange)
                foreach (double unmatchedUwomen in unmatchedUtilRange)
                {
                    utilFunction.setCoefficient(RandomTraitsLineUtil.MEN_UNMATCHED_UTIL, unmatchedUmen);
                    utilFunction.setCoefficient(RandomTraitsLineUtil.WOMEN_UNMATCHED_UTIL, unmatchedUwomen);

                    for (int numMen = 20; numMen <= 60; numMen++)
                        runAndPrintForParams(utilFunction, sw, numMen, numWomen, iterationNum, isFullStats);

                    Console.WriteLine("Short Lists \t{0:n}\t {1:n}", unmatchedUmen, unmatchedUwomen);
                }
            sw.Close();
        }




        /// Utility Scripts and function from here below
        //////////////////////////////////////////////
        //////////////////////////////////////////////
        //////////////////////////////////////////////




        /// <summary>
        /// This funcation takes the utility model and parameteres, runs the requested number of simulations and writes the results to the stream writer.
        /// Rank is converted to 1-base (i.e., most prefered partner is rank =1). 
        /// 
        /// If utilFunction is null the function will draw uniform IID preferences for all agents.  
        /// </summary>
        /// <param name="utilFunction"></param>
        /// <param name="sw"></param>
        /// <param name="numMen"></param>
        /// <param name="numWomen"></param>
        /// <param name="iterationNum"></param>
        /// <param name="isFullStats"></param>
        public static void runAndPrintForParams(IUtility utilFunction, StreamWriter sw, int numMen, int numWomen, int iterationNum, bool isFullStats)
        {

            // Setup of variables
            Economy e = new Economy(numMen, numWomen);

            double[] rankMOSMArr = new double[iterationNum];
            double[] rankWOSMArr = new double[iterationNum];
            int[] numMatchedArr = new int[iterationNum];
            int[] numMultArr = new int[iterationNum];




            // run iterations for the given parameters
            for (int iteration = 0; iteration < iterationNum; iteration++)
            {
                if (utilFunction != null)
                {
                    e.drawPrefFromUtility(utilFunction);
                }
                else
                {
                    e.drawUniformIIDPref();
                }

                e.calcMatchingStats();

                // Calculating rank, and converting from 0-base (best is rank 0) to 1-base (best is rank 1).
                rankMOSMArr[iteration] = (e.totalRankMOSM / ((double)e.numMatched)) + 1;
                rankWOSMArr[iteration] = (e.totalRankWOSM / ((double)e.numMatched)) + 1;
                numMatchedArr[iteration] = e.numMatched;
                numMultArr[iteration] = e.numMult;

            }

            // print out stats for these param

            sw.Write("#Men=\t{0}\t #Women=\t{1}\t", numMen, numWomen);


            if (utilFunction != null)
                utilFunction.writeCoefficients(sw);


            sw.Write("R(M)_MOSM\t");
            printStats(rankMOSMArr, sw, isFullStats);
            sw.Write("R(M)_WOSM\t");
            printStats(rankWOSMArr, sw, isFullStats);
            sw.Write("PctMenMult\t");
            printStats(numMultArr, ((double)numMen) / 100.0, sw, isFullStats);
            sw.Write("PctMenMatched\t");
            printStats(numMatchedArr, ((double)numMen) / 100.0, sw, isFullStats);
            sw.WriteLine();

        }


        public static void printAvg(int[] arr, StreamWriter sw)
        {
            double avg = 0;
            for (int i = 0; i < arr.Length; i++)
                avg += arr[i];
            avg = avg / arr.Length;
            sw.Write("{0:f3}\t", avg);
        }


        public static void printStats(int[] inputArray, double normalizationFactor, StreamWriter sw, bool isFullStats)
        {
            double avg = 0;
            double std = 0;
            double pct95;
            double pct5;

            double[] arr = new double[inputArray.Length];

            for (int i = 0; i < arr.Length; i++)
            {
                arr[i] = ((double)inputArray[i]) / normalizationFactor;
            }

            for (int i = 0; i < arr.Length; i++)
            {
                avg += arr[i];
            }
            avg = avg / arr.Length;
            sw.Write("{0:f2}\t", avg);

            if (isFullStats)
            {


                for (int i = 0; i < arr.Length; i++)
                {
                    std += (avg - arr[i]) * (avg - arr[i]);
                }
                std = Math.Sqrt(std / arr.Length);

                Array.Sort(arr);

                pct95 = arr[(int)Math.Round(0.95 * (arr.Length - 1))];
                pct5 = arr[(int)Math.Round(0.05 * (arr.Length - 1))];

                sw.Write("{0:f2}\t", std);
                sw.Write("{0:f2}\t", pct95);
                sw.Write("{0:f2}\t", pct5);
            }

            /*
            sw.Write("avg = {0:f3}\t", avg);
            sw.Write("std = {0:f3}\t", std);
            sw.Write("pct95 = {0:f3}\t", pct95);
            sw.Write("pct5 = {0:f3}\t", pct5);
             */
        }


        public static void printStats(double[] arr, StreamWriter sw, bool isFullStats)
        {
            double avg = 0;


            for (int i = 0; i < arr.Length; i++)
            {
                avg += arr[i];
            }
            avg = avg / arr.Length;
            sw.Write("{0:f3}\t", avg);

            if (isFullStats)
            {

                double std = 0;
                double pct95;
                double pct5;

                for (int i = 0; i < arr.Length; i++)
                {
                    std += (avg - arr[i]) * (avg - arr[i]);
                }
                std = Math.Sqrt(std / arr.Length);

                Array.Sort(arr);

                pct95 = arr[(int)Math.Round(0.95 * (arr.Length - 1))];
                pct5 = arr[(int)Math.Round(0.05 * (arr.Length - 1))];

                sw.Write("{0:f3}\t", std);
                sw.Write("{0:f3}\t", pct95);
                sw.Write("{0:f3}\t", pct5);
            }
            /*
            sw.Write("avg = {0:f3}\t", avg);
            sw.Write("std = {0:f3}\t", std);
            sw.Write("pct95 = {0:f3}\t", pct95);
            sw.Write("pct5 = {0:f3}\t", pct5);
             */
        }


    }

}

