﻿// RandomTraitsLineUtility.cs
// Class representing a ranodm utility with verticle and horizontal preferences

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UnbalancedMatchSim;

namespace UnbalancedMatchSim.Utilities
{
    /// <summary>
    /// Interface that holds characteristics of agents
    /// </summary>
    public class DiffAbsCharac : ICharacteristics
    {

        public const string DIFF_ATTR_NAME = "diffAttr";
        public const string ABS_ATTR_NAME = "absAttr";

        public const string PREFERENCE_DIRECTION_ATTR = "prefDirection";

        Dictionary<string, double[]> charData = new Dictionary<string, double[]>();
        int K;


        /// <summary>
        /// Inits the attributes data by making a double array of the given size for each attribute. 
        /// </summary>
        /// <param name="K"></param>
        public DiffAbsCharac(int K)
        {
            this.K = K;
            charData.Add(DIFF_ATTR_NAME, new double[K]);
            charData.Add(ABS_ATTR_NAME, new double[K]);
            charData.Add(PREFERENCE_DIRECTION_ATTR, new double[K]);
        }

        public double getCharacteristic(int agentIndex, string attribute)
        {
            return charData[attribute][agentIndex];
        }

        public void setCharacteristic(int agentIndex, string attribute, double value)
        {
            charData[attribute][agentIndex] = value;
        }

        public void printCharac(StreamWriter sw)
        {
            sw.WriteLine("\nAttributes: " + ABS_ATTR_NAME  + " " + DIFF_ATTR_NAME + " " + PREFERENCE_DIRECTION_ATTR);
            for (int i = 0; i < K; i++)
            {
                sw.Write(i + "\t");
                sw.WriteLine("\t{0:n} \t{1:n} \t{2:F0}", getCharacteristic(i, ABS_ATTR_NAME),
                    getCharacteristic(i, DIFF_ATTR_NAME),
                    getCharacteristic(i, PREFERENCE_DIRECTION_ATTR)
                    );
            }
            sw.WriteLine();
        }
    }

    /// <summary>
    /// This class implements a utility function that gives each agent an absolute attribute and a differential attribute.
    /// Both attributes are drawn U[0,1]
    /// The differential attribute can in the segment [0,1], giving utility of the form:
    /// u_i(j) = \beta x_j^A - \gamma^+ (| x_i^D - x_j^D|_+ )^2 - \gamma^- (| x_j^D - x_i^D|_+ )^2 + \varepsilon_{ij},
    /// 
    /// </summary>
    public class RandomTraitsLineUtil : IUtility
    {
        public const string UTIL_LABELROW = "xUtil	coeAbs	coeDiffMin	coeDiffPlus	coeExp	uBarM	uBarW	coeRand	";
        


        public const string COE_DIFF_PLUS = "coeDiffPlus";
        public const string COE_DIFF_MIN = "coeDiffMin";
        public const string COE_ABS = "coeAbs";
        public const string COE_RANDOM_DRAW = "randomDrawCoe";
        public const string COE_DIFF_EXPONENET = "diffExponent";

        public const string MEN_UNMATCHED_UTIL = "uBarM";
        public const string WOMEN_UNMATCHED_UTIL = "uBarW";

        const double PLUS_VALUE = +1;
        const double MINUS_VALUE = -1;



        // If this flag is true the coefficient on the diff attribute will be 
        //  multiplied by -1 for women
        public const string PARAM_IS_DIFF_REVERSED_FOR_WOMEN = "diffReversedWomen";

        public const int PARAM_TRUE = 1;
        public const int PARAM_FALSE = 0;



        protected Dictionary<string, double> coefficients = new Dictionary<string, double>();
        protected Dictionary<string, int> parameters = new Dictionary<string, int>();

        /// <summary>
        /// Inits the random utility with all coeficients set to 0
        /// </summary>
        public RandomTraitsLineUtil()
        {
            coefficients[COE_ABS] = 0;
            coefficients[COE_DIFF_MIN] = 0;
            coefficients[COE_DIFF_PLUS] = 0;
            coefficients[COE_RANDOM_DRAW] = 1;
            coefficients[COE_DIFF_EXPONENET] = 1;

            coefficients[MEN_UNMATCHED_UTIL] = double.NegativeInfinity;
            coefficients[WOMEN_UNMATCHED_UTIL] = double.NegativeInfinity;
         
            parameters[PARAM_IS_DIFF_REVERSED_FOR_WOMEN] = PARAM_FALSE;
        }



        public void setCoefficient(string coefficientName, double value)
        {
            coefficients[coefficientName] = value;
        }

        public void setParam(string paramName, int value)
        {
            parameters[paramName] = value;
        }


        public static Random r = new Random();


        public static double getExremeValDraw()
        {
            return -Math.Log(-Math.Log(r.NextDouble()));
        }


        public ICharacteristics drawChar(int K, bool isMen)
        {
            if (isMen)
                return drawMenChar(K);
            else
                return drawWomenChar(K);
        }

        /// <summary>
        /// Draws charectaristics for men. 
        /// Sets both ABS and Diff to be a uniform [0,1]
        /// </summary>
        /// <param name="K"></param>
        /// <returns></returns>
        public ICharacteristics drawMenChar(int K)
        {
            DiffAbsCharac charMen = new DiffAbsCharac(K);

            for (int i = 0; i < K; i++)
            {
                charMen.setCharacteristic(i, DiffAbsCharac.ABS_ATTR_NAME, r.NextDouble());
                charMen.setCharacteristic(i, DiffAbsCharac.DIFF_ATTR_NAME, r.NextDouble());
                charMen.setCharacteristic(i, DiffAbsCharac.PREFERENCE_DIRECTION_ATTR, PLUS_VALUE);
            }

            return charMen;

        }

        public ICharacteristics drawWomenChar(int K)
        {
            DiffAbsCharac charWomen = new DiffAbsCharac(K);

            for (int i = 0; i < K; i++)
            {
                charWomen.setCharacteristic(i, DiffAbsCharac.ABS_ATTR_NAME, r.NextDouble());
                charWomen.setCharacteristic(i, DiffAbsCharac.DIFF_ATTR_NAME, r.NextDouble());
                // set the preference direction to be opposite of men, if the paramter indicates so
                if (parameters[PARAM_IS_DIFF_REVERSED_FOR_WOMEN] == PARAM_TRUE)
                {
                    charWomen.setCharacteristic(i, DiffAbsCharac.PREFERENCE_DIRECTION_ATTR, MINUS_VALUE);
                }
                else
                {
                    charWomen.setCharacteristic(i, DiffAbsCharac.PREFERENCE_DIRECTION_ATTR, PLUS_VALUE);
                }
            }

            return charWomen;
        }


        public double getUnmatchedUtil(int agentIndex, ICharacteristics characteristicsSpouses, ICharacteristics mySideCharestaristics, bool isMen)
        {
            if (isMen)
                return coefficients[MEN_UNMATCHED_UTIL];
            else
                return coefficients[WOMEN_UNMATCHED_UTIL];

        }


        public double drawUtil(int agentIndex, int spouseIndex, ICharacteristics characteristicsSpouses, ICharacteristics charestaristicsMySide, bool isMen)
        {
            double absAtt = characteristicsSpouses.getCharacteristic(spouseIndex, DiffAbsCharac.ABS_ATTR_NAME);

            double diffPrefDirection = charestaristicsMySide.getCharacteristic(agentIndex, DiffAbsCharac.PREFERENCE_DIRECTION_ATTR);

            double diff = charestaristicsMySide.getCharacteristic(agentIndex, DiffAbsCharac.DIFF_ATTR_NAME)
                          - characteristicsSpouses.getCharacteristic(spouseIndex, DiffAbsCharac.DIFF_ATTR_NAME);
            // Utility model when attributes are on the line [0,1]
            //////////////////////////////////////////////////////

            double diffPlus = Math.Pow( Math.Max(0, diff),  coefficients[COE_DIFF_EXPONENET]);
            double diffMinus = Math.Pow(Math.Max(0, -diff), coefficients[COE_DIFF_EXPONENET]);


            double u = diffPlus * coefficients[COE_DIFF_PLUS] * diffPrefDirection
                       + diffMinus * coefficients[COE_DIFF_MIN] * diffPrefDirection
                       + absAtt * coefficients[COE_ABS]
                       + PrefGenerator.getExremeValDraw() * coefficients[COE_RANDOM_DRAW];
            return u;
        }


        public void writeCoefficients(StreamWriter sw)
        {
            sw.Write("UtilCoe\t");
            sw.Write("{0:n}\t", coefficients[COE_ABS]);
            sw.Write("{0:n}\t", coefficients[COE_DIFF_MIN]);
            sw.Write("{0:n}\t", coefficients[COE_DIFF_PLUS]);
            sw.Write("{0:n}\t", coefficients[COE_DIFF_EXPONENET]);
            
            sw.Write("{0:n}\t", coefficients[MEN_UNMATCHED_UTIL]);
            sw.Write("{0:n}\t", coefficients[WOMEN_UNMATCHED_UTIL]);
            
            sw.Write("{0:n}\t", coefficients[COE_RANDOM_DRAW]);
        }

        public void writeUtilityModelDescrition(StreamWriter sw)
        {
            
            sw.WriteLine(" utility model: \n u(x) = c1* x1 + c2 * (y2 - x2)_^a+ + c3 * (x2 - y2)_^a+ + c4 * eps");
            sw.WriteLine(" Where my char are (y1,y2) and the spouse is (x1,x2), and eps is an extreme value draw.");
            if (parameters[PARAM_IS_DIFF_REVERSED_FOR_WOMEN] == PARAM_TRUE)
                sw.WriteLine(" The coeficient c2 is replaced by -c2 for women");
            sw.Write("Coefficients printed in following order: c1 = " + COE_ABS + ",\t");
            sw.Write("c2 = " + COE_DIFF_MIN + ",\t");
            sw.Write("c3 = " + COE_DIFF_PLUS + ",\t");
            sw.Write("c4 = " + COE_RANDOM_DRAW + "\t");
            sw.Write("a  = " + COE_DIFF_EXPONENET + "\t");
            sw.WriteLine();
        }
    }
}
