﻿// UnequalMatching.cs
// This class includes the function MPDA (Deferred Acceptance) and MOSM2WOSM (our new algorithm)

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.IO;

namespace UnbalancedMatchSim
{


    /// <summary>
    /// This class includes functions that calculate stable matchings. 
    /// The function MPDA can be used to calculate both MOSM or the WOSM (since men and women are symmetric).
    /// 
    /// The function MOSM2WOSM implements our new matching algorithm, and efficently computes the WOSM for unbalanced 
    /// markets with stritcly more women than men. It can be run only after running MPDA for men. 
    /// </summary>
    class UnequalMatching
    {
      

        /// <summary>
        /// Assumees that MPDA was run already and the agents hold assignments that constitute a stable matching.
        /// Using the algorithm we have in the paper.
        /// Assumes that K>N
        /// (number of women > number of men)
        /// </summary>
        /// <param name="men"></param>
        /// <param name="women"></param>
        /// <returns></returns>
        public static int[] MOSM2WOSM(Agent[] men, Agent[] women)
        {
            int N = men.Length;
            int K = women.Length;

            SortedSet<int> S = new SortedSet<int>();
            SortedSet<int> WminS = new SortedSet<int>();

            int[] muTilde = new int[K];
            int[] muTildeM = new int[N];

            // Start by putting in S all unmatched women
            // and init muTilde to be the woman's initial match
            for (int j = 0; j < K; j++)
            {
                if (women[j].isUnmatched())
                    S.Add(j);
                else
                    WminS.Add(j);
                muTilde[j] = women[j].getAssignedSpouse();
            }

            // and save a muTildeM to save men's assignment
            for (int i = 0; i < N; i++)
            {
                muTildeM[i] = men[i].getAssignedSpouse();
            }




            while (WminS.Count > 0)
            {
                ArrayList V = new ArrayList();
                ArrayList Vmen = new ArrayList();

                int wHat = WminS.First();

                //Divorce wHat: reject her husband, but keep her "assigned rank" which will be compared to future proposals

                V.Add(wHat);
                debugPr("\n\nwHat=" + wHat + "");

                Agent proposer = men[women[wHat].getAssignedSpouse()];
                proposer.reject();
                Vmen.Add(proposer.ID);

                Boolean inRejectionChain = true;
                Agent target = women[proposer.getAssignedSpouse()];
                // the proposer proposes and one of the four possible things happen

                while (inRejectionChain)
                {
                    debugPr("\n* p=m" + proposer.ID + "\t" + "t=w" + target.ID + "\t");
                    bool stepDone = false;

                    // (I) & (II)                    
                    if (!S.Contains(target.ID) && !V.Contains(target.ID) && !stepDone)
                    {
                        debugPr("1+2\t");
                        stepDone = true;
                        // proposal goes to target woman
                        int rejectedManID = target.makeProposal(proposer.ID);

                        if (rejectedManID == proposer.ID)
                        {
                            // (I)
                            debugPr("1 ");
                            debugPr(" R:m" + rejectedManID + " ");
                            proposer.reject();
                            debugPr(" T:w" + proposer.getAssignedSpouse() + " ");
                            target = women[proposer.getAssignedSpouse()];
                        }
                        else
                        {
                            // (II)
                            debugPr("2");
                            // (Since the target woman is not in S she was rejected another man)
                            V.Add(target.ID);
                            proposer = men[rejectedManID];
                            proposer.reject();
                            Vmen.Add(rejectedManID);
                            target = women[proposer.getAssignedSpouse()];
                        }
                    }


                    // (I) & (III)
                    if (V.Contains(target.ID) && !stepDone)
                    {
                        debugPr("1+3\t");
                        stepDone = true;
                        // check if the current man is better than the orignal match under muTilde
                        if (target.getRank(proposer.ID) < target.getRank(muTilde[target.ID]))
                        {
                            // (III)
                            debugPr("3");
                            // we need to update muTilde for all women in the improvment cycle
                            int vIndex = V.IndexOf(target.ID);
                            int[] Varr = (int[])V.ToArray(typeof(int));
                            int[] VmenArr = (int[])Vmen.ToArray(typeof(int));

                            debugPr(" vInd=" + vIndex);

                            // the woman that closed the cycle gets the proposing man
                            muTilde[target.ID] = proposer.ID;
                            muTildeM[proposer.ID] = target.ID;

                            // the rest of women in the cycle get the men they are currenly assigned 
                            // (and likewise for the men)
                            for (int j = vIndex + 1; j < V.Count; j++)
                            {
                                int wID = Varr[j];
                                muTilde[wID] = women[wID].getAssignedSpouse();
                                int mID = VmenArr[j];
                                muTildeM[mID] = men[mID].getAssignedSpouse();
                            }
                            V.RemoveRange(vIndex, V.Count - vIndex);
                            Vmen.RemoveRange(vIndex, Vmen.Count - vIndex);


                            if (vIndex > 0)
                            {

                                // finally, if vIndex>0, that is, the woman is in the middle of the chain,
                                // the proposer will be the man currently assigned to the woman who closed the cycle
                                // he will propose again to her, as she got a new husband from the cycle

                                Agent nextProposer = men[target.getAssignedSpouse()];
                                target.setUnassigned();
                                target.makeProposal(muTilde[target.ID]);


                                //proposer.setUnassigned();
                                //proposer.makeProposal(muTildeM[proposer.ID]);

                                proposer = nextProposer;
                            }
                            else
                            {
                                //otherwise, assign the proposer to wHat 
                                //we can end the rejction chain and start a new phase
                                target.makeProposal(proposer.ID);
                                inRejectionChain = false;
                            }

                        }
                        else
                        {

                            // (I)
                            debugPr("1");
                            proposer.reject();
                            target = women[proposer.getAssignedSpouse()];

                        }


                    }





                    // (I) & (IV)
                    if (S.Contains(target.ID) && !stepDone)
                    {
                        debugPr("1+4\t");
                        if (target.isUnmatched() || target.getRank(proposer.ID) < target.getRank(muTilde[target.ID]))
                        {
                            debugPr("4");
                            stepDone = true;
                            //add all the women from the chain to S
                            foreach (int w in V)
                            {
                                S.Add(w);
                                WminS.Remove(w);
                            }
                            inRejectionChain = false;
                        }
                        else
                        {
                            // (I)
                            debugPr("1");
                            proposer.reject();
                            target = women[proposer.getAssignedSpouse()];

                        }
                    }

                }
            }

            return muTilde;
        }







        public static int[] MPDA(Agent[] men, Agent[] women)
        {
            int N = men.Length;
            int K = women.Length;



            // Set all women to be unmatched
            for (int j = 0; j < K; j++)
                women[j].setUnassigned();

            int activeMenCount = 0;

            while (activeMenCount < N)
            {
                Agent proposer = men[activeMenCount];

                proposer.setTopSpouse();
                if (Agent.UNMATCHED_INDICATOR != proposer.getAssignedSpouse())
                {
                    Agent target = women[proposer.getAssignedSpouse()];
                    int rejectedAgentID = target.makeProposal(proposer.ID);


                    while (Agent.UNMATCHED_INDICATOR != rejectedAgentID)
                    {
                        proposer = men[rejectedAgentID];
                        proposer.reject(); // make the proposer point to the next woman in his pref
                        // This may create a conflict if this woman already has a proposal

                        if (proposer.isUnmatched())
                        {
                            // if the men got beyond the end of his list, he remains unmatched
                            rejectedAgentID = Agent.UNMATCHED_INDICATOR;
                        }
                        else
                        {
                            // the men poposer to the next woman on his list, and may displace another man
                            target = women[proposer.getAssignedSpouse()];
                            rejectedAgentID = target.makeProposal(proposer.ID);
                        }
                    }
                }
                activeMenCount++;
            }

            int[] matching = new int[N];

            for (int i = 0; i < N; i++)
            {
                matching[i] = men[i].getAssignedSpouse();
            }
            return matching;
        }

        
        /// Code for debugging
        //////////////////////


        public static bool DEBUG_PRINTS = false;
        // public static StreamWriter sw = new StreamWriter("./Output/RunOutput.txt", false, Encoding.ASCII);
        public static StreamWriter sw = null;

        static void debugPr(string s)
        {
            if (DEBUG_PRINTS)
                sw.Write(s);
        }

        static void printMatch(int[] mu)
        {
            if (DEBUG_PRINTS)
            {
                sw.WriteLine();
                for (int i = 0; i < mu.Length; i++)
                {
                    sw.WriteLine(" " + i + " " + mu[i]);
                }
                sw.WriteLine("\n--\n");
            }
        }

    }
}
